MAIN_CONTROLLER: str = "__main__"

LEFT_FRONT_WHEEL_SPEED_SENSOR: str = "wss_lf"
RIGHT_FRONT_WHEEL_SPEED_SENSOR: str = "wss_rf"
CENTER_REAR_WHEEL_SPEED_SENSOR: str = "wss_cr"
LEFT_REAR_WHEEL_SPEED_SENSOR: str = "wss_lr"
RIGHT_REAR_WHEEL_SPEED_SENSOR: str = "wss_rr"
ACCELEROMETER: str = "accl"
ODOMETER: str = "odm"
VOLTAGE_SENSOR: str = "vot"
THROTTLE_PEDAL: str = "tpd"
BRAKE_PEDAL: str = "bpd"
MOTOR_CONTROLLER: str = "mc"
BRAKE_CONTROLLER: str = "bc"
GPS_RECEIVER: str = "gps"
LEFT_INDICATOR: str = "lindi"
RIGHT_INDICATOR: str = "rindi"
